///////////////////////////////////////////////////////////////////////////////
// File: Connection.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2009 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "Connection.h"
#include "IManageListeningConnections.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/ErrorCodeToErrorMessage.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/AddressRenderer.h"
#include "JetByteTools/SocketTools/ICreateStreamSocketConnections.h"
#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/SmartStreamSocket.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::OutputEx;
using JetByteTools::Core::ErrorCodeToErrorMessageA;
using JetByteTools::Core::ILockableObject;

using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;

using JetByteTools::Socket::ICreateStreamSocketConnections;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::CAddressRenderer;
using JetByteTools::Socket::CFullAddress;
using JetByteTools::Socket::CSmartStreamSocket;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static CFullAddress ParseConnectionString(
   const string &connectionString);

///////////////////////////////////////////////////////////////////////////////
// CConnection
///////////////////////////////////////////////////////////////////////////////

CConnection::CConnection(
   IStreamSocket &socket,
   ICreateStreamSocketConnections &connectionCreator,
   IManageListeningConnections &connectionManager)
   :  m_localSocket(socket),
      m_pRemoteSocket(nullptr),
      m_connectionCreator(connectionCreator),
      m_connectionManager(connectionManager),
      m_ref(1),
      m_state(Waiting)
{

}

void CConnection::AddRef()
{
   ::InterlockedIncrement(&m_ref);
}

void CConnection::Release()
{
   const unsigned long ref = ::InterlockedDecrement(&m_ref);

   if (ref == 0)
   {
      m_connectionManager.RemoveConnection(*this);

      delete this;
   }
}

void CConnection::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress &address)
{
   ILockableObject::Owner lock(m_lock);

   if (&socket == &m_localSocket)
   {
      DisplayCommandPrompt();
   }
   else
   {
      if (m_pRemoteSocket)
      {
         throw CException(_T("CConnection::OnConnectionEstablished()"), _T("Remote socket already connected."));
      }

      m_pRemoteSocket = &socket;

      if (m_state == Listening)
      {
         AddRef();

         SendMessage("Inbound connection established from: " + CStringConverter::TtoA(CAddressRenderer::AsString(address, true)));

         const string message = "\r\nConnected\r\n";

         m_pRemoteSocket->TryWrite(message.c_str(), static_cast<DWORD>(message.length()));
      }
      else
      {
         SendMessage("Outbound connection established to: " + CStringConverter::TtoA(CAddressRenderer::AsString(address, true)));
      }

      m_state = Connected;
   }

   socket.TryRead();
}

void CConnection::OnConnectionFailed(
   IStreamSocket &/*socket*/,
   const IAddress &address,
   const DWORD lastError)
{
   SendMessage("Outbound connection to: " + CStringConverter::TtoA(CAddressRenderer::AsString(address, true)) + " failed: " + ErrorCodeToErrorMessageA(lastError));

   m_state = Waiting;

   DisplayCommandPrompt();

   Release();
}

void CConnection::DisplayCommandPrompt() const
{
   const string message = "\r\nReady: [D]isconnect, [L]isten or [C]onnect xxx.xxx.xxx.xxx:port ?\r\n";

   SendMessage(message);
}

void CConnection::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   ILockableObject::Owner lock(m_lock);

   if (&socket == &m_localSocket)
   {
      OnLocalReadCompleted(buffer);
   }
   else if (&socket == m_pRemoteSocket)
   {
      OnRemoteReadCompleted(buffer);
   }
   else
   {
      throw CException(_T("CConnection::OnConnectionClosed()"), _T("Unexpected socket."));
   }
}

void CConnection::OnConnectionClosed(
   IStreamSocket &socket)
{
   ILockableObject::Owner lock(m_lock);

   if (&socket == &m_localSocket)
   {
      m_state = Disconnected;

      if (m_pRemoteSocket)
      {
         m_pRemoteSocket->Shutdown();
      }
   }
   else if (&socket == m_pRemoteSocket)
   {
      m_pRemoteSocket = nullptr;

      m_state = Waiting;

      DisplayCommandPrompt();
   }
   else
   {
      throw CException(_T("CConnection::OnConnectionClosed()"), _T("Unexpected socket."));
   }
}

void CConnection::OnLocalReadCompleted(
   CSmartBuffer &buffer)
{
   if (m_pRemoteSocket)
   {
      m_pRemoteSocket->TryWrite(buffer);

      m_localSocket.TryRead();
   }
   else
   {
      switch (m_state)
      {
         case Waiting :

            ProcessData(buffer);

         break;

         case Disconnected :

            OutputEx(_T("Data recieved on socket that has disconnected"));

         break;

         case Listening :

            SendErrorAndDisconnect("Connection is currently listening, we cannot accept further commands in this state");

         break;

         case Connecting :

            SendErrorAndDisconnect("Connection is currently connecting, we cannot accept further commands in this state");

         break;

         case Connected :

            SendErrorAndDisconnect("Unexpected: Connection is connected but remote socket is null");

         break;
      }
   }
}

void CConnection::SendErrorAndDisconnect(
   const string &message)
{
   SendMessage(message);

   m_localSocket.Shutdown();
}

void CConnection::SendMessage(
   const string &message) const
{
   const auto fullMessage = "\r\n" + message + "\r\n";

   m_localSocket.TryWrite(fullMessage.c_str(), static_cast<DWORD>(fullMessage.length()));
}

void CConnection::ProcessData(
   CSmartBuffer &buffer)
{
   OutputEx(DumpData(buffer->GetMemory(), buffer->GetUsed()));

   auto done = false;

   while (!done)
   {
      auto foundReturn = false;

      const auto pData = reinterpret_cast<const char *>(buffer->GetMemory());

      const auto dataLength = buffer->GetUsed();

      const auto finalCharacter = dataLength - 1;

      for (IBuffer::BufferSize i = 0; i < dataLength; ++i)
      {
         if (pData[i] == '\r')
         {
            foundReturn = true;
         }
         else if (pData[i] == '\n' && foundReturn)
         {
            if (i == finalCharacter)
            {
               ProcessCommand(buffer);

               m_localSocket.TryRead();

               done = true;
            }
            else
            {
               auto command(buffer->SplitBufferFromFront(i + 1));

               ProcessCommand(command);
            }
         }
         else if (i == finalCharacter)
         {
            m_localSocket.TryRead(buffer);

            done = true;
         }
      }
   }
}

void CConnection::ProcessCommand(
   CSmartBuffer &buffer)
{
   buffer->AddData(0);   // null terminate the command string

   const auto pData = reinterpret_cast<const char *>(buffer->GetMemory());

   if (*pData == 'D' || *pData == 'd')
   {
      SendMessage("Good bye.");

      m_localSocket.Shutdown();
   }
   else if (*pData == 'L' || *pData == 'l')
   {
      OnListen();
   }
   else if (*pData == 'C' || *pData == 'c')
   {
      OnConnect(pData);
   }
   else
   {
      SendMessage("Unrecognised command: \"" + string(pData, buffer->GetUsed() - 3) + "\" expected D, L or C xxx.xxx.xxx.xxx:port \r\n");

      DisplayCommandPrompt();
   }
}

void CConnection::OnListen()
{
   SendMessage("Listening for inbound connections.");

   m_state = Listening;

   m_connectionManager.AddConnection(*this);
}

void CConnection::OnConnect(
   const string &connectionString)
{
   const auto address = ParseConnectionString(connectionString);

   SendMessage("Connecting to \"" + CStringConverter::TtoA(CAddressRenderer::AsString(address, true)) + "\".");

   m_state = Connecting;

   AddRef();

   auto socket(m_connectionCreator.AsyncConnectNoThrow(address, this));

   if (!socket.Get())
   {
      throw CException(
         _T("CConnection::OnConnect()"),
         _T("Failed to connect to remote server at: ") + CAddressRenderer::AsString(address, true));
   }
}

void CConnection::OnRemoteReadCompleted(
   CSmartBuffer &buffer)
{
   m_localSocket.TryWrite(buffer);

   m_pRemoteSocket->TryRead();
}

///////////////////////////////////////////////////////////////////////////////
// Static helper functions
///////////////////////////////////////////////////////////////////////////////

static CFullAddress ParseConnectionString(
   const string &connectionString)
{
   // Connection string is in one of these formats:
   // "C xxx.xxx.xxx.xxx:port\r\n"
   // "C [ffff:ffff:ffff:ffff:ffff]:port\r\n"

   static const size_t commandLength = 2; //"C "

   static const size_t crlfLength = 2;    // \r\n

   const auto address = connectionString.substr(commandLength, connectionString.length() - (commandLength + crlfLength));

   return CFullAddress(CStringConverter::AtoT(address));
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
