///////////////////////////////////////////////////////////////////////////////
// File: InboundSocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2009 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "InboundSocketServer.h"
#include "Connection.h"
#include "IProvideListeningConnections.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"

#pragma hdrstop

#include "JetByteTools/SocketTools/IAllocateStreamSockets.h"

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::Output;
using JetByteTools::Core::CException;
using JetByteTools::Core::IIndexedOpaqueUserData;

using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IIOPool;

using JetByteTools::Socket::CStreamSocketServerEx;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ILimitConnections;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::NoZeroByteRead;

///////////////////////////////////////////////////////////////////////////////
// CInboundSocketServer
///////////////////////////////////////////////////////////////////////////////

CInboundSocketServer::CInboundSocketServer(
   IProvideListeningConnections &connectionProvider,
   const IFullAddress &address,
   const ListenBacklog listenBacklog,
   IIOPool &pool,
   IAllocateStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator,
   ILimitConnections &connectionLimiter)
   :  CStreamSocketServerEx(address, listenBacklog, *this, pool, socketAllocator, bufferAllocator, NoZeroByteRead, connectionLimiter),
      m_userDataIndex(socketAllocator.RequestUserDataSlot(_T("CInboundSocketServer - m_userDataIndex"))),
      m_connectionProvider(connectionProvider)
{

}

CInboundSocketServer::~CInboundSocketServer()
{
   WaitForShutdownToComplete();
}

void CInboundSocketServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress &address)
{
   Output(_T("OnConnectionEstablished"));

   // look up a listening connection from the collection, if we find one, wire this up to
   // it, else close the connection...

   auto pConnection = m_connectionProvider.GetListeningConnection();

   if (pConnection)
   {
      socket.SetUserPointer(m_userDataIndex, pConnection);

      pConnection->OnConnectionEstablished(socket, address);
   }
}

void CInboundSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   try
   {
      auto pConnection = reinterpret_cast<CConnection *>(socket.GetUserPointer(m_userDataIndex));

      pConnection->OnReadCompleted(socket, buffer);
   }
   catch(const CException &e)
   {
      Output(_T("ReadCompleted - Exception - ") + e.GetDetails());
      socket.AbortConnection();
   }
   catch(...)
   {
      Output(_T("ReadCompleted - Unexpected exception"));
      socket.AbortConnection();
   }
}

void CInboundSocketServer::OnConnectionClosed(
   IStreamSocket &socket)
{
   Output(_T("OnConnectionClosed"));

   auto pConnection = reinterpret_cast<CConnection *>(socket.GetUserPointer(m_userDataIndex));

   if (pConnection)
   {
      pConnection->OnConnectionClosed(socket);
   }
}

void CInboundSocketServer::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   Output(_T("OnSocketReleased"));

   auto pConnection = reinterpret_cast<CConnection *>(userData.GetUserPointer(m_userDataIndex));

   if (pConnection)
   {
      pConnection->Release();
   }
}

///////////////////////////////////////////////////////////////////////////////
// End of file: InboundSocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
