///////////////////////////////////////////////////////////////////////////////
// File: Connection.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2009 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"
#include "JetByteTools/Admin/WinsockWrapper.h"

#include "Connection.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/AddressRenderer.h"
#include "JetByteTools/SocketTools/FullAddress.h"
#include "JetByteTools/SocketTools/SmartStreamSocket.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::_tstring;
using JetByteTools::Core::CException;
using JetByteTools::Core::ILockableObject;

using JetByteTools::IO::CSmartBuffer;

using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::CSmartStreamSocket;

///////////////////////////////////////////////////////////////////////////////
// Connection
///////////////////////////////////////////////////////////////////////////////

CConnection::CConnection(
   IStreamSocket &socket)
   :  m_localSocket(socket),
      m_pRemoteSocket(nullptr),
      m_ref(1)
{

}

void CConnection::AddRef()
{
   ::InterlockedIncrement(&m_ref);
}

void CConnection::Release()
{
   const unsigned long ref = ::InterlockedDecrement(&m_ref);

   if (ref == 0)
   {
      delete this;
   }
}

void CConnection::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   ILockableObject::Owner lock(m_lock);

   if (&socket != &m_localSocket)
   {
      if (m_pRemoteSocket)
      {
         throw CException(_T("CConnection::OnConnectionEstablished()"), _T("Remote socket already connected."));
      }

      m_pRemoteSocket = &socket;

      AddRef();

   }

   socket.TryRead();
}

void CConnection::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   ILockableObject::Owner lock(m_lock);

   if (&socket == &m_localSocket)
   {
      OnLocalReadCompleted(buffer);
   }
   else if (&socket == m_pRemoteSocket)
   {
      OnRemoteReadCompleted(buffer);
   }
   else
   {
      throw CException(_T("CConnection::OnConnectionClosed()"), _T("Unexpected socket."));
   }
}

void CConnection::OnConnectionClosed(
   IStreamSocket &socket)
{
   ILockableObject::Owner lock(m_lock);

   if (&socket == &m_localSocket)
   {
      if (m_pRemoteSocket)
      {
         m_pRemoteSocket->Shutdown();
      }
   }
   else if (&socket == m_pRemoteSocket)
   {
      m_pRemoteSocket = nullptr;

      m_localSocket.Shutdown();
   }
   else
   {
      throw CException(_T("CConnection::OnConnectionClosed()"), _T("Unexpected socket."));
   }
}

void CConnection::OnLocalReadCompleted(
   CSmartBuffer &buffer)
{
   if (m_pRemoteSocket)
   {
      m_pRemoteSocket->Write(buffer);

      m_localSocket.TryRead();
   }
}

void CConnection::OnRemoteReadCompleted(
   CSmartBuffer &buffer)
{
   m_localSocket.TryWrite(buffer);

   m_pRemoteSocket->TryRead();
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
